import numpy as np

from autolens.data.fitting.util import fitting_util
from autolens.model.inversion import inversions


class AbstractFit(object):

    def __init__(self, fitting_datas, model_datas_):
        """Abstract base class of a fit to a dataset.

        All fitting quantities (datas, model_datas, residuals, etc.) are stored as lists, where the different entries
        of the list correspond to different data-sets. To signify this, all quantities are pluralized. However, if a non-
        plural term is used (e.g. residual) the residuals of the first regular are returned. This makes the interface
        for fitting single data-sets (the most common use-case) more intuitive.

        A fit is performed using masked data which has been mapped to 1D arrays, even if the data is originally a 2D
        regular. The 1D representation of a data is followed by an underscore. For example, model_datas gives the 2D
        images of the data, whereas model_datas_ gives the 1D data representation.

        Parameters
        -----------
        fitting_datas : [fitting_data.FittingImage]
            The fitting images that are fitted.
        model_datas_ : [ndarray]
            The model datas the fitting images are fitted with.

        Attributes
        -----------
        datas : [scaled_array.ScaledSquarePixelArray]
            The data that is fitted.
        noise_maps : [scaled_array.ScaledSquarePixelArray]
            The noise-maps used to fit the data.
        masks : [regular.Mask]
            The masks which define the regions of the data that are fitted.
        model_datas : [scaled_array.ScaledSquarePixelArray]
            The model dataset used to fit the data
        map_to_scaled_arrays : [func]
            Functions which map the masked 1D arrays of the data, model_data, residuals, etc. to their 2D data
            representation (if an regular)
        residuals : [scaled_array.ScaledSquarePixelArray]
            The residuals of the fit (data - model_data).
        chi_squareds : [scaled_array.ScaledSquarePixelArray]
            The chi-sqaureds of the fit ((data - model_data) / noise_map ) **2.0
        chi_squared_terms : [float]
            The summed chi-squared of every data-point in a fit.
        chi_squareds_term : float
            The summed chi_squared_terms for all datasets.
        reduced_chi_squared_terms : [float]
            The reduced chi_squared of the fit to every data-set (chi_squared / number of data points).
        reduced_chi_squared_term : [float]
            The summed reduced chi_squared of the fit between data and model for all datasets.
        noise_terms : [float]
            The normalization term of a likelihood function assuming Gaussian noise in every data-point.
        noise_term : float
            The sum of all noise_terms for all datasets.
        likelihoods : [float]
            The likelihood of every fit between data and model -0.5 * (chi_squared_term + noise_term)
        likelihood : float
            The summed likelihood of the fit between data and model for all datasets.
        """

        self.datas_ = list(map(lambda fit_data : fit_data[:], fitting_datas))
        self.noise_maps_ = list(map(lambda fit_data : fit_data.noise_map_, fitting_datas))
        self.masks = list(map(lambda fit_data : fit_data.mask, fitting_datas))
        self.map_to_scaled_arrays = list(map(lambda fit_data: fit_data.grids.regular.scaled_array_from_array_1d,
                                             fitting_datas))

        self.model_datas_ = model_datas_
        self.residuals_ = fitting_util.residuals_from_datas_and_model_datas(datas_=self.datas_, 
                                                                            model_datas_=self.model_datas_)
        self.chi_squareds_ = fitting_util.chi_squareds_from_residuals_and_noise_maps(self.residuals_, self.noise_maps_)

    @property
    def chi_squared_terms(self):
        return fitting_util.chi_squared_terms_from_chi_squareds(self.chi_squareds_)

    @property
    def chi_squared_term(self):
        return sum(self.chi_squared_terms)

    @property
    def reduced_chi_squareds(self):
        return list(map(lambda chi_squared_term, mask : chi_squared_term / mask.pixels_in_mask,
                        self.chi_squared_terms, self.masks))

    @property
    def reduced_chi_squared(self):
        return sum(self.reduced_chi_squareds)

    @property
    def noise_terms(self):
        return fitting_util.noise_terms_from_noise_maps(self.noise_maps_)

    @property
    def noise_term(self):
        return sum(self.noise_terms)

    @property
    def likelihoods(self):
        return fitting_util.likelihoods_from_chi_squareds_and_noise_terms(self.chi_squared_terms, self.noise_terms)

    @property
    def likelihood(self):
        return sum(self.likelihoods)

    @property
    def noise_maps(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.noise_maps_, map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def model_datas(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.model_datas_, map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def residuals(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.residuals_, map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def chi_squareds(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.chi_squareds_, map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def noise_map(self):
        return self.noise_maps[0]

    @property
    def model_data(self):
        return self.model_datas[0]

    @property
    def residual(self):
        return self.residuals[0]

    @property
    def chi_squared(self):
        return self.chi_squareds[0]


class AbstractDataFit(AbstractFit):

    def __init__(self, fitting_datas, model_datas_):
        """Abstract Base class of a fit between data-set and model, see *AbstractFit*.

        Parameters
        -----------
        fitting_datas : [fitting_data.FittingImage]
            The fitting images that are fitted.
        model_datas_ : [ndarray]
            The model datas the fitting images are fitted with."""

        self.datas = list(map(lambda fit_data : fit_data.array, fitting_datas))
        super(AbstractDataFit, self).__init__(fitting_datas=fitting_datas, model_datas_=model_datas_)


class AbstractImageFit(AbstractFit):

    def __init__(self, fitting_images, model_images_):
        """Abstract Base class of a fit between data-set and model, see *AbstractFit*.

        This class is used to add the terms 'images' and 'model_images' to the attributes of a fit. However, they
        function identical to datas and model_datas, and this is done onoly for the user interface.

        Parameters
        -----------
        fitting_images : [fitting_data.FittingImage]
            The fitting images that are fitted.
        model_datas_ : [ndarray]
            The model images the fitting images are fitted with.
        """

        self.images = list(map(lambda fit_image : fit_image.image, fitting_images))
        super(AbstractImageFit, self).__init__(fitting_datas=fitting_images, model_datas_=model_images_)

    @property
    def model_images(self):
        return self.model_datas

    @property
    def model_image(self):
        return self.model_datas[0]


class AbstractHyperFit(object):

    def __init__(self, fitting_hyper_images, hyper_galaxies):
        """Abstract base class of a hyper-fit.

        A hyper-fit is a fit which performs a fit as described in the *AbstractFit*, but also includes a set of
        parameters which allow the noise-map of the data-set to be scaled. This is done to prevent over-fitting
        small regions of a data-set with high chi-squared values and therefore provide a global fit to the overall
        data-set.

        This is performed using an existing model of the data-set to compute a contributions regular, which a set of
        hyper-parameters then use to increase the noise in localized regions of the data-set.

        Parameters
        -----------
        fitting_hyper_images : [fitting_data.FittingHyperImage]
            The fitting images that are fitted, which include the hyper-images used for scaling the noise-map.
        hyper_galaxies : [galaxy.Galaxy]
            The hyper-galaxies which represent the model components used to scale the noise, which correspond to
            individual galaxies in the regular.

        Attributes
        -----------
        contributions : [[scaled_array.ScaledSquarePixelArray]]
            The contribution map of every regular, where there is an individual contribution map for each hyper-galaxy in
            the model.
        scaled_noise_maps : [scaled_array.ScaledSquarePixelArray]
            The scaled noise maps of the regular, computed after using the hyper-galaxies.
        scaled_chi_squared_terms : [float]
            The summed scaled chi-squared of every data-point in a fit.
        scaled_chi_squareds_term : float
            The sum of all scaled_chi_squared_terms for all images.
        scaled_noise_terms : [float]
            The normalization term of a likelihood function assuming Gaussian noise in every data-point, using the
            scaled noise-map.
        scaled_noise_term : float
            The sum of all scaled_noise_terms for all images.
        scaled_likelihoods : [float]
            The likelihood of every fit between data and model using the scaled noise-map's fit \
            -0.5 * (scaled_chi_squared_term + scaled_noise_term)
        scaled_likelihood : float
            The summed scaled likelihood of the fit between data and model for all images.
        """

        self.is_hyper_fit = True
        self.contributions_ = \
            fitting_util.contributions_from_fitting_hyper_images_and_hyper_galaxies(fitting_hyper_images=fitting_hyper_images,
                                                                       hyper_galaxies=hyper_galaxies)


        self.scaled_noise_maps_ =\
            fitting_util.scaled_noises_from_fitting_hyper_images_contributions_and_hyper_galaxies(
                fitting_hyper_images=fitting_hyper_images, contributions_=self.contributions_,
                hyper_galaxies=hyper_galaxies)

    @property
    def scaled_chi_squared_terms(self):
        return fitting_util.chi_squared_terms_from_chi_squareds(self.scaled_chi_squareds_)

    @property
    def scaled_noise_terms(self):
        return fitting_util.noise_terms_from_noise_maps(self.scaled_noise_maps_)

    @property
    def scaled_noise_maps(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.scaled_noise_maps_,
                                           map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def scaled_chi_squareds(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.scaled_chi_squareds_,
                                           map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def contributions(self):
        contributions = [[] for _ in range(len(self.contributions_))]
        for image_index in range(len(contributions)):
            contributions[image_index] = list(map(lambda _contributions :
                                                  self.map_to_scaled_arrays[image_index](_contributions),
                                                  self.contributions_[image_index]))
        return contributions

    @property
    def scaled_noise_map(self):
        return self.scaled_noise_maps[0]

    @property
    def scaled_chi_squared(self):
        return self.scaled_chi_squareds[0]


class AbstractHyperImageFit(AbstractImageFit, AbstractHyperFit):

    def __init__(self, fitting_hyper_images, model_images_, hyper_galaxies):
       """Abstract base class for an regular data-set which includes hyper noise-scaling. Seee *AbstractFit* and
       *AbstractHyperFit* for more details."""
       AbstractHyperFit.__init__(self=self, fitting_hyper_images=fitting_hyper_images, hyper_galaxies=hyper_galaxies)
       super(AbstractHyperImageFit, self).__init__(fitting_images=fitting_hyper_images, model_images_=model_images_)
       self.scaled_chi_squareds_ = fitting_util.chi_squareds_from_residuals_and_noise_maps(self.residuals_, self.scaled_noise_maps_)


class AbstractConvolutionFit(AbstractImageFit):

    def __init__(self, fitting_images, images_, blurring_images_):
        """Abstract base class for a fit to an regular using a light-profile.

        This includes the blurring of the light-profile model regular with the instrumental PSF.

        Parameters
        -----------
        fitting_images : [fitting_data.FittingImage]
            The fitting images that are fitted.
        images_ : [ndarray]
            The masked 1D representation of the light profile model images before PSF blurring.
        blurring_images_ : [ndarray]
            The 1D representation of the light profile model images blurring region, which corresponds to all pixels \
            which are not inside the mask but close enough that their light will be blurred into it via PSF convolution.
        """

        self.convolvers_image = list(map(lambda fit_image : fit_image.convolver_image, fitting_images))

        model_images_ = fitting_util.blur_images_including_blurring_regions(images_=images_, blurring_images_=blurring_images_,
                                                               convolvers=self.convolvers_image)

        super(AbstractConvolutionFit, self).__init__(fitting_images=fitting_images, model_images_=model_images_)


class AbstractInversionFit(AbstractImageFit):

    def __init__(self, fitting_images, mapper, regularization):
        """Abstract base class for a fit to an regular which uses a linear inversion.

        This includes passing the regular / noise_map / PSF and inversion objects to the inversons module to perform \
        the inversion.

        Parameters
        -----------
        fitting_images : [fitting_data.FittingImage]
            The fitting images that are fitted.
        mapper : inversion.mapper.Mapper
            Class storing the mappings between observed regular-pixels and inversion's pixelization pixels.
        regularization : inversion.regularization.Regularization
            Class storing the regularization scheme of the inversion's pixelization.

        Attributes
        -----------
        likelihoods_with_regularization : [float]
            The likelihood of each fit to the regular, including jsut 3 terms, the chi-squared term, regularization
            penalty factor and noise normalization term.
        likelihood_with_regularization : float
            The sum of all likelihoods_with_regularization for all images.
        evidences : [float]
            The Bayesian evidence of each fit to the regular. The Bayesian evidence is described in Suyu et al. 2006 and
            the howtolens inversion tutorial chapter_4_inversions/tutorial_4_bayesian_regularization.
        evidence : float
            The sum of evidence values for all images.
        """

        self.inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(image=fitting_images[0][:],
                         noise_map=fitting_images[0].noise_map_, convolver=fitting_images[0].convolver_mapping_matrix,
                                                                         mapper=mapper, regularization=regularization)

        super(AbstractInversionFit, self).__init__(fitting_images=fitting_images,
                                                   model_images_=[self.inversion.reconstructed_data_vector])

    @property
    def likelihoods_with_regularization(self):
        return fitting_util.likelihoods_with_regularization_from_chi_squared_regularization_and_noise_terms(self.chi_squared_terms,
               [self.inversion.regularization_term], self.noise_terms)

    @property
    def likelihood_with_regularization(self):
        return sum(self.likelihoods_with_regularization)

    @property
    def evidences(self):
        return fitting_util.evidences_from_reconstruction_terms(self.chi_squared_terms, 
                                                                [self.inversion.regularization_term],
                                                               [self.inversion.log_det_curvature_reg_matrix_term],
                                                               [self.inversion.log_det_regularization_matrix_term],
                                                               self.noise_terms)

    @property
    def evidence(self):
        return sum(self.evidences)


class AbstractConvolutionInversionFit(AbstractImageFit):
    
    def __init__(self, fitting_images, images_, blurring_images_, mapper, regularization):
        
        self.convolvers_image = list(map(lambda fit_image : fit_image.convolver_image, fitting_images))

        self.profile_model_images_ = fitting_util.blur_images_including_blurring_regions(images_=images_,
                                                                            blurring_images_=blurring_images_,
                                                                            convolvers=self.convolvers_image)

        self.profile_subtracted_images_ = list(map(lambda fitting_image, profile_model_image_ :
                                                   fitting_image[:] - profile_model_image_,
                                                   fitting_images, self.profile_model_images_))

        self.inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(
            image=self.profile_subtracted_images_[0], noise_map=fitting_images[0].noise_map_,
            convolver=fitting_images[0].convolver_mapping_matrix, mapper=mapper, regularization=regularization)
        
        self.inversion_model_images_ = [self.inversion.reconstructed_data_vector]

        model_images_ = list(map(lambda profile_model_image_, inversion_model_image_ :
                                 profile_model_image_ + inversion_model_image_,
                                 self.profile_model_images_, self.inversion_model_images_))

        super(AbstractConvolutionInversionFit, self).__init__(fitting_images=fitting_images, model_images_=model_images_)

    @property
    def profile_subtracted_images(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.profile_subtracted_images_,
                                           map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def profile_model_images(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.profile_model_images_,
                                           map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def inversion_model_images(self):
        return fitting_util.map_arrays_to_scaled_arrays(arrays_=self.inversion_model_images_,
                                           map_to_scaled_arrays=self.map_to_scaled_arrays)

    @property
    def profile_subtracted_image(self):
        return self.profile_subtracted_images[0]

    @property
    def profile_model_image(self):
        return self.profile_model_images[0]

    @property
    def inversion_model_image(self):
        return self.inversion_model_images[0]

    @property
    def evidences(self):
        return fitting_util.evidences_from_reconstruction_terms(self.chi_squared_terms, [self.inversion.regularization_term],
                                                   [self.inversion.log_det_curvature_reg_matrix_term],
                                                   [self.inversion.log_det_regularization_matrix_term],
                                                   self.noise_terms)

    @property
    def evidence(self):
        return sum(self.evidences)



import inspect

from autofit import exc
from autofit.core import model_mapper
from autofit.core.model_mapper import PriorNameValue, ConstantNameValue, cast_collection

from autolens.model.galaxy import galaxy
from autolens.model.profiles import light_profiles, mass_profiles


def is_light_profile_class(cls):
    """
    Parameters
    ----------
    cls
        Some object

    Returns
    -------
    bool: is_light_profile_class
        True if cls is a class that inherits from light profile
    """
    return inspect.isclass(cls) and issubclass(cls, light_profiles.LightProfile)


def is_mass_profile_class(cls):
    """
    Parameters
    ----------
    cls
        Some object

    Returns
    -------
    bool: is_mass_profile_class
        True if cls is a class that inherits from mass profile
    """
    return inspect.isclass(cls) and issubclass(cls, mass_profiles.MassProfile)


def is_profile_class(cls):
    """
    Parameters
    ----------
    cls
        Some object

    Returns
    -------
    bool: is_mass_profile_class
        True if cls is a class that inherits from light profile or mass profile
    """
    return is_light_profile_class(cls) or is_mass_profile_class(cls)


class GalaxyModel(model_mapper.AbstractPriorModel):
    """
    @DynamicAttrs
    """

    @property
    def flat_prior_model_tuples(self):
        return [flat_prior_model for prior_model in self.prior_models for flat_prior_model in
                prior_model.flat_prior_model_tuples]

    def __init__(self, align_centres=False, align_axis_ratios=False, align_orientations=False, redshift=None,
                 variable_redshift=False, pixelization=None, regularization=None, hyper_galaxy=None, **kwargs):
        """Class to produce Galaxy instances from sets of profile classes and other model-fitting attributes (e.g. \
         pixelizations, regularization schemes, hyper-galaxyes) using the model mapper.

        Parameters
        ----------
        light_profile_classes: [LightProfile]
            The *LightProfile* classes for which model light profile instances are generated for this galaxy model.
        mass_profile_classes: [MassProfile]
            The *MassProfile* classes for which model mass profile instances are generated for this galaxy model.
        align_centres : bool
            If *True*, the same prior will be used for all the profiles centres, such that all light and / or mass \
            profiles always share the same origin.
        align_axis_ratios : bool
            If *True*, the same prior will be used for all the profiles axis-ratio, such that all light and / or mass \
            profiles always share the same axis-ratio.
        align_orientations : bool
            If *True*, the same prior will be used for all the profiles rotation angles phi, such that all light \
            and / or mass profiles always share the same orientation.
        redshift : float
            The redshift of this model galaxy.
        variable_redshift : bool
            If *True*, the galaxy redshift will be treated as a free-parameter that is fitted for by the non-linear \
            search.
        pixelization : Pixelization
            The pixelization used to reconstruct the galaxy light and fit the observed regular if using an inversion.
        regularization : Regularization
            The regularization-scheme used to regularization reconstruct the galaxy light when fitting the observed \
            regular if using an inversion.
        hyper_galaxy : HyperGalaxy
            A model hyper-galaxy used for scaling the observed regular's noise.
        """

        self.align_centres = align_centres
        self.align_axis_ratios = align_axis_ratios
        self.align_orientations = align_orientations

        profile_models = []

        for name, cls in kwargs.items():
            if is_mass_profile_class(cls) or is_light_profile_class(cls):
                model = model_mapper.PriorModel(cls)
                profile_models.append(model)
                setattr(self, name, model)
            else:
                setattr(self, name, cls)

        if len(profile_models) > 0:
            if self.align_centres:
                centre = profile_models[0].centre
                for profile_model in profile_models:
                    profile_model.centre = centre

            if self.align_axis_ratios:
                axis_ratio = profile_models[0].axis_ratio
                for profile_model in profile_models:
                    profile_model.axis_ratio = axis_ratio

            if self.align_orientations:
                phi = profile_models[0].phi
                for profile_model in profile_models:
                    profile_model.phi = phi

        if redshift is not None:
            self.redshift = model_mapper.Constant(
                redshift.redshift if isinstance(redshift, galaxy.Redshift) else redshift)
        else:
            self.redshift = model_mapper.PriorModel(galaxy.Redshift) if variable_redshift else model_mapper.Constant(1)

        if pixelization is not None and regularization is None:
            raise exc.PriorException('If the galaxy prior has a pixelization, it must also have a regularization.')
        if pixelization is None and regularization is not None:
            raise exc.PriorException('If the galaxy prior has a regularization, it must also have a pixelization.')

        self.pixelization = model_mapper.PriorModel(pixelization) if inspect.isclass(
            pixelization) else pixelization
        self.regularization = model_mapper.PriorModel(regularization) if inspect.isclass(
            regularization) else regularization

        self.hyper_galaxy = model_mapper.PriorModel(hyper_galaxy) if inspect.isclass(hyper_galaxy) else hyper_galaxy

    def __setattr__(self, key, value):
        if key == "redshift" \
                and (isinstance(value, float)
                     or isinstance(value, int)
                     or isinstance(value, model_mapper.Prior)
                     or isinstance(value, model_mapper.Constant)):
            value = galaxy.Redshift(value)
        super(GalaxyModel, self).__setattr__(key, value)

    def linked_model_for_classes(self, **classes):
        light_profile_class_tuples = [(name, cls) for name, cls in classes.items() if is_light_profile_class(cls)]
        mass_profile_class_tuples = [(name, cls) for name, cls in classes.items() if is_mass_profile_class(cls)]

        def link_models_to_classes(models, class_tuples):
            result = []
            for n, class_tuple in enumerate(class_tuples):
                result.append((class_tuple[0], models[n % len(models)].linked_model_for_class(class_tuple[1])))
            return result

        light_profile_model_tuples = link_models_to_classes(self.light_profile_prior_models, light_profile_class_tuples)
        mass_profile_model_tuples = link_models_to_classes(self.mass_profile_prior_models, mass_profile_class_tuples)

        return GalaxyModel(**{**dict(light_profile_model_tuples), **dict(mass_profile_model_tuples)})

    @property
    def constant_light_profiles(self):
        """
        Returns
        -------
        light_profiles: [light_profiles.LightProfile]
            Light profiles with set variables
        """
        return [value for value in self.__dict__.values() if galaxy.is_light_profile(value)]

    @property
    def constant_mass_profiles(self):
        """
        Returns
        -------
        mass_profiles: [mass_profiles.MassProfile]
            Mass profiles with set variables
        """
        return [value for value in self.__dict__.values() if galaxy.is_mass_profile(value)]

    @property
    def prior_models(self):
        """
        Returns
        -------
        prior_models: [model_mapper.PriorModel]
            A list of the prior models (e.g. variable profiles) attached to this galaxy prior
        """
        return [value for _, value in
                filter(lambda t: isinstance(t[1], model_mapper.PriorModel), self.__dict__.items())]

    @property
    def profile_prior_model_dict(self):
        """
        Returns
        -------
        profile_prior_model_dict: {str: PriorModel}
            A dictionary mapping_matrix instance variable names to variable profiles.
        """
        return {key: value for key, value in
                filter(lambda t: isinstance(t[1], model_mapper.PriorModel) and is_profile_class(t[1].cls),
                       self.__dict__.items())}

    @property
    def light_profile_prior_models(self):
        return [item for item in self.__dict__.values() if
                isinstance(item, model_mapper.PriorModel) and is_light_profile_class(item.cls)]

    @property
    def mass_profile_prior_models(self):
        return [item for item in self.__dict__.values() if
                isinstance(item, model_mapper.PriorModel) and is_mass_profile_class(item.cls)]

    @property
    def constant_profile_dict(self):
        """
        Returns
        -------
        constant_profile_dict: {str: geometry_profiles.GeometryProfile}
            A dictionary mapping_matrix instance variable names to profiles with set variables.
        """
        return {key: value for key, value in self.__dict__.items() if
                galaxy.is_light_profile(value) or galaxy.is_mass_profile(value)}

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples(self):
        """
        Returns
        -------
        priors: [PriorTuple]
            A list of priors associated with prior models in this galaxy prior.
        """
        return [prior for prior_model in self.prior_models for prior in prior_model.prior_tuples]

    @property
    @cast_collection(ConstantNameValue)
    def constant_tuples(self):
        """
        Returns
        -------
        constant: [ConstantTuple]
            A list of constants associated with prior models in this galaxy prior.
        """
        return [constant for prior_model in self.prior_models for constant in prior_model.constant_tuples]

    @property
    def prior_class_dict(self):
        """
        Returns
        -------
        prior_class_dict: {Prior: class}
            A dictionary mapping_matrix priors to the class associated with their prior model.
        """
        return {prior: cls for prior_model in self.prior_models for prior, cls in
                prior_model.prior_class_dict.items()}

    def instance_for_arguments(self, arguments):
        """
        Create an instance of the associated class for a set of arguments

        Parameters
        ----------
        arguments: {Prior: value}
            Dictionary mapping_matrix priors to attribute analysis_path and value pairs

        Returns
        -------
            An instance of the class
        """
        profiles = {**{key: value.instance_for_arguments(arguments)
                       for key, value
                       in self.profile_prior_model_dict.items()}, **self.constant_profile_dict}

        if isinstance(self.redshift, galaxy.Redshift):
            redshift = self.redshift
        else:
            redshift = self.redshift.instance_for_arguments(arguments)
        pixelization = self.pixelization.instance_for_arguments(arguments) \
            if isinstance(self.pixelization, model_mapper.PriorModel) \
            else self.pixelization
        regularization = self.regularization.instance_for_arguments(arguments) \
            if isinstance(self.regularization, model_mapper.PriorModel) \
            else self.regularization
        hyper_galaxy = self.hyper_galaxy.instance_for_arguments(arguments) \
            if isinstance(self.hyper_galaxy, model_mapper.PriorModel) \
            else self.hyper_galaxy

        return galaxy.Galaxy(redshift=redshift.redshift, pixelization=pixelization, regularization=regularization,
                             hyper_galaxy=hyper_galaxy, **profiles)

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Create a new galaxy prior from a set of arguments, replacing the priors of some of this galaxy prior's prior
        models with new arguments.

        Parameters
        ----------
        arguments: dict
            A dictionary mapping_matrix between old priors and their replacements.

        Returns
        -------
        new_model: GalaxyModel
            A model with some or all priors replaced.
        """
        new_model = GalaxyModel(align_centres=self.align_centres, align_axis_ratios=self.align_axis_ratios,
                                align_orientations=self.align_orientations)

        for key, value in filter(lambda t: isinstance(t[1], model_mapper.PriorModel), self.__dict__.items()):
            setattr(new_model, key, value.gaussian_prior_model_for_arguments(arguments))

        return new_model

    @classmethod
    def from_galaxy(cls, g, **kwargs):
        """
        Create a new galaxy prior with constants taken from a galaxy.

        Parameters
        ----------
        g: galaxy.Galaxy
            A galaxy
        kwargs
            Key word arguments to override GalaxyModel constructor arguments.
        """
        return GalaxyModel(**{**g.__dict__, **kwargs})



import numpy as np
import numba

@numba.jit(nopython=True, parallel=True)
def mapping_matrix_from_sub_to_pix(sub_to_pix, pixels, regular_pixels, sub_to_regular, sub_grid_fraction):
    """Computes the mapping matrix, by iterating over the known mappings between the sub-grid and pixelization.

    Parameters
    -----------
    sub_to_pix : ndarray
        The mappings between the observed regular's sub-pixels and pixelization's pixels.
    pixels : int
        The number of pixels in the pixelization.
    regular_pixels : int
        The number of image pixels in the observed image and thus on the regular grid.
    sub_to_regular : ndarray
        The mappings between the observed regular's sub-pixels and observed regular's pixels.
    sub_grid_fraction : float
        The fractional area each sub-pixel takes up in an regular-pixel.
    """

    mapping_matrix = np.zeros((regular_pixels, pixels))

    for sub_index in range(sub_to_regular.shape[0]):
        mapping_matrix[sub_to_regular[sub_index], sub_to_pix[sub_index]] += sub_grid_fraction

    return mapping_matrix

@numba.jit(nopython=True, parallel=True)
def voronoi_regular_to_pix_from_grids_and_geometry(regular_grid, regular_to_nearest_regular_pix, pixel_centres,
                                                   pixel_neighbors, pixel_neighbors_size):
    """ Compute the mappings between a set of sub-maskedimage pixels and pixels, using the maskedimage's traced \
    pix-plane sub-grid and the pixel centers. This uses the pix-neighbors to perform a graph \
    search when pairing pixels, for efficiency.

    For the Voronoi pixelizations, a cluster set of 'cluster-pixels' are used to determine the pixelization. \
    These provide the mappings between only a sub-set of sub-pixels / maskedimage-pixels and pixels.

    To determine the complete set of sub-pixel to pixel mappings, we must therefore pair every sub-pixel to \
    its nearest pixel (using the sub-pixel's pix-plane coordinate and pixel center). Using a full \
    nearest neighbor search to do this is slow, thus the pixel neighbors (derived via the Voronoi grid) \
    is used to localize each nearest neighbor search.

    In this routine, some variables and function names refer to a 'cluster_pix_'. This term describes a \
    pixel that we have paired to a sub_coordinate using the cluster_coordinate of an maskedimage coordinate. \
    Thus, it may not actually be that sub_coordinate's closest pixel (the routine will eventually
    determine this).

    Parameters
    ----------
    coordinate : [float, float]
        The x and y pix sub-grid grid which are to be matched with their closest pixels.
    nearest_pixel : int
        The nearest pixel defined on the cluster-pixel grid.
    cluster_to_pixelization : [int]
        The mapping_matrix between every cluster-pixel and pixel (e.g. if the fifth pixel maps to \
        the 3rd cluster_pixel, cluster_to_pix[4] = 2).
     """

    regular_to_pix = np.zeros((regular_grid.shape[0]))

    for regular_index in range(regular_grid.shape[0]):

        nearest_pix_pixel_index = regular_to_nearest_regular_pix[regular_index]

        while True:

            nearest_pix_pixel_center = pixel_centres[nearest_pix_pixel_index]

            sub_to_nearest_pix_distance = (regular_grid[regular_index, 0] - nearest_pix_pixel_center[0]) ** 2 + \
                                          (regular_grid[regular_index, 1] - nearest_pix_pixel_center[1]) ** 2

            closest_separation_from_pix_neighbor = 1.0e8

            for neighbor_index in range(pixel_neighbors_size[nearest_pix_pixel_index]):

                neighbor = pixel_neighbors[nearest_pix_pixel_index, neighbor_index]

                separation_from_neighbor = (regular_grid[regular_index, 0] - pixel_centres[neighbor, 0]) ** 2 + \
                                           (regular_grid[regular_index, 1] - pixel_centres[neighbor, 1]) ** 2

                if separation_from_neighbor < closest_separation_from_pix_neighbor:
                    closest_separation_from_pix_neighbor = separation_from_neighbor
                    closest_neighbor_index = neighbor_index

            neighboring_pix_pixel_index = pixel_neighbors[nearest_pix_pixel_index, closest_neighbor_index]
            sub_to_neighboring_pix_distance = closest_separation_from_pix_neighbor

            if sub_to_nearest_pix_distance <= sub_to_neighboring_pix_distance:
                regular_to_pix[regular_index] = nearest_pix_pixel_index
                break
            else:
                nearest_pix_pixel_index = neighboring_pix_pixel_index

    return regular_to_pix

@numba.jit(nopython=True, parallel=True)
def voronoi_sub_to_pix_from_grids_and_geometry(sub_grid, regular_to_nearest_regular_pix, sub_to_regular, pixel_centres,
                                               pixel_neighbors, pixel_neighbors_size):
    """ Compute the mappings between a set of sub-maskedimage pixels and pixels, using the maskedimage's traced \
    pix-plane sub-grid and the pixel centers. This uses the pix-neighbors to perform a graph \
    search when pairing pixels, for efficiency.

    For the Voronoi pixelizations, a cluster set of 'cluster-pixels' are used to determine the pixelization. \
    These provide the mappings between only a sub-set of sub-pixels / maskedimage-pixels and pixels.

    To determine the complete set of sub-pixel to pixel mappings, we must therefore pair every sub-pixel to \
    its nearest pixel (using the sub-pixel's pix-plane coordinate and pixel center). Using a full \
    nearest neighbor search to do this is slow, thus the pixel neighbors (derived via the Voronoi grid) \
    is used to localize each nearest neighbor search.

    In this routine, some variables and function names refer to a 'cluster_pix_'. This term describes a \
    pixel that we have paired to a sub_coordinate using the cluster_coordinate of an maskedimage coordinate. \
    Thus, it may not actually be that sub_coordinate's closest pixel (the routine will eventually
    determine this).

    Parameters
    ----------
    coordinate : [float, float]
        The x and y pix sub-grid grid which are to be matched with their closest pixels.
    nearest_pixel : int
        The nearest pixel defined on the cluster-pixel grid.
    cluster_to_pixelization : [int]
        The mapping_matrix between every cluster-pixel and pixel (e.g. if the fifth pixel maps to \
        the 3rd cluster_pixel, cluster_to_pix[4] = 2).
     """

    sub_to_pix = np.zeros((sub_grid.shape[0]))

    for sub_index in range(sub_grid.shape[0]):

        nearest_pix_pixel_index = regular_to_nearest_regular_pix[sub_to_regular[sub_index]]

        while True:

            nearest_pix_pixel_center = pixel_centres[nearest_pix_pixel_index]

            sub_to_nearest_pix_distance = (sub_grid[sub_index, 0] - nearest_pix_pixel_center[0]) ** 2 + \
                                          (sub_grid[sub_index, 1] - nearest_pix_pixel_center[1]) ** 2

            closest_separation_from_pix_to_neighbor = 1.0e8

            for neighbor_index in range(pixel_neighbors_size[nearest_pix_pixel_index]):

                neighbor = pixel_neighbors[nearest_pix_pixel_index, neighbor_index]

                separation_from_neighbor = (sub_grid[sub_index, 0] - pixel_centres[neighbor, 0]) ** 2 + \
                                           (sub_grid[sub_index, 1] - pixel_centres[neighbor, 1]) ** 2

                if separation_from_neighbor < closest_separation_from_pix_to_neighbor:
                    closest_separation_from_pix_to_neighbor = separation_from_neighbor
                    closest_neighbor_index = neighbor_index

            neighboring_pix_pixel_index = pixel_neighbors[nearest_pix_pixel_index, closest_neighbor_index]
            sub_to_neighboring_pix_distance = closest_separation_from_pix_to_neighbor

            if sub_to_nearest_pix_distance <= sub_to_neighboring_pix_distance:
                sub_to_pix[sub_index] = nearest_pix_pixel_index
                break
            else:
                nearest_pix_pixel_index = neighboring_pix_pixel_index

    return sub_to_pix


from autolens.data.imaging import image as im
from autolens.data.array import grids, mask as ma
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lensing import ray_tracing
from autolens.lensing import lensing_image as li
from autolens.model.inversion import pixelizations as pix
from autolens.data.imaging.plotters import imaging_plotters
from autolens.model.inversion.plotters import mapper_plotters


# In the previous example, we used a mapper to make a rectangular pixelization. However, it wasn't clear what a mapper
# was actually mapping. Infact, it didn't do much mapping at all! Therefore, in this tutorial, we'll cover mapping.

# To begin, lets simulate and load an regular - it'll be clear why we're doing this in a moment.
def simulate():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lensing import ray_tracing

    psf = im.PSF.simulate_as_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    image_plane_grids = grids.ImagingGrids.grids_for_simulation(shape=(150, 150), pixel_scale=0.05, psf_shape=(11, 11))

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0,
                                                        einstein_radius=1.6))
    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.7, phi=135.0, intensity=0.2,
                                                       effective_radius=0.2, sersic_index=2.5))
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grids=[image_plane_grids])

    return im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.05,
                                        exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

image = simulate()
imaging_plotters.plot_image_subplot(image=image)

# Lets begin by setting up our grids (using the regular we loaded above).
image_plane_grids = grids.DataGrids.from_shape_and_pixel_scale(shape=image.shape, pixel_scale=image.pixel_scale,
                                                               sub_grid_size=2)

# Our tracer will use the same lens model_galaxy and source model_galaxy that we used to simulate the regular (although, becuase
# we're modeling the source with a pixel-grid, we don't need to supply its light profile).
lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0,
                                                    einstein_radius=1.6))
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[g.Galaxy()],
                                             image_plane_grids=[image_plane_grids])

# Finally, lets setup our pixelization and mapper, like we did before, using the tracer's source-plane grid.
rectangular = pix.Rectangular(shape=(25, 25))
mapper = rectangular.mapper_from_grids_and_border(grids=tracer.source_plane.grids[0], border=None)

# Again, we're going to plot our mapper, but we're also going to plot the regular which was used to generate the grid we
# mapped to the source-plane.
mapper_plotters.plot_image_and_mapper(image=image, mapper=mapper, should_plot_grid=True)

# The pixels in the regular map to the pixels in the source-plane, and visa-versa. Lets highlight a set of regular-pixels
# in both the regular and source-plane.
mapper_plotters.plot_image_and_mapper(image=image, mapper=mapper, should_plot_grid=True,
                                      image_pixels=[[range(0, 100)], [range(900, 1000)]])

# That's nice, and we can see the mappings, but it isn't really what we want to know, is it? We really want to go the
# other way, and see how our source-pixels map to the regular. This is where mappers come into their own, as they let us
# map all the points in a given source-pixel back to the regular. Lets map source pixel 313, the central
# source-pixel, to the regular.
mapper_plotters.plot_image_and_mapper(image=image, mapper=mapper, should_plot_grid=True,
                                      source_pixels=[[312]])

# And there we have it - multiple imaging in all its glory. Try changing the source-pixel indexes of the line below.
# This will give you a feel for how different regions of the source-plane map to the regular.
mapper_plotters.plot_image_and_mapper(image=image, mapper=mapper, should_plot_grid=True,
                                      source_pixels=[[312, 318], [412]])

# Okay, so I think we can agree, mappers map things! More specifically, they map our source-plane pixelization to an
# observed regular of a strong lens.
#
# Finally, lets do the same as above, but using a masked regular. By applying a masks, the mapper will only map
# regular-pixels inside the masks. This removes the (many) regular pixels at the edge of the regular, where the source clearly
# isn't present and which pad-out the size of the source-plane. Lets just have a quick look at these edges pixels:
mapper_plotters.plot_image_and_mapper(image=image, mapper=mapper, should_plot_grid=True,
                                      source_pixels=[[0, 1, 2, 3, 4, 5, 6, 7], [620, 621, 622, 623]])

# Lets use an annular masks, which will capture the ring-like shape of the lensed source model_galaxy.
mask = ma.Mask.circular_annular(shape=image.shape, pixel_scale=image.pixel_scale, inner_radius_arcsec=1.0,
                                outer_radius_arcsec=2.2)

# Lets quickly confirm the annuli radii capture the source's light
#imaging_plotters.plot_image(regular=regular, masks=masks)

# As usual, we setup our regular and masks up as a lensing regular and create a tracer using its (now masked) grids.
lensing_image = li.LensingImage(image=image, mask=mask)
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[g.Galaxy()],
                                             image_plane_grids=[lensing_image.grids])

# Finally, we use the tracer's (masked) source-plane grid to setup a new mapper (using the same rectangular 25 x 25
# pixelization as before).
mapper = rectangular.mapper_from_grids_and_border(grids=tracer.source_plane.grids[0], border=None)

# Lets have another look
mapper_plotters.plot_image_and_mapper(image=image, mask=mask, mapper=mapper, should_plot_grid=True)

# Woah! Look how much closer we are to the source-plane (The axis sizes have decreased from ~ -2.5" -> 2.5" to
# ~ -0.6" to 0.6"). We can now really see the diamond of points in the origin of the source-plane (for those who have
# been reading up, this diamond is called the 'caustic'). This diamond defines when lensing moves from the quadruply
# imaged regime to doubly-imaged regime, and we can actually show this now using our mapper now.
mapper_plotters.plot_image_and_mapper(image=image, mask=mask, mapper=mapper, should_plot_grid=True,
                                      source_pixels=[[312], [314], [316], [318]])

# Great - tutorial 2 down! We've learnt about mappers, which map things, and we used them to understand how the regular
# and source plane map to one another. Your exercises are:

# 1) Change the einstein radius of the lens model_galaxy in small increments (e.g. einstein radius 1.6" -> 1.55").
#    As the radius deviates from 1.6" (the input value of the simulated lens), what do you notice about where the
#    points map from the origin of the source-plane (where the source-model_galaxy is simulated, e.g. (0.0", 0.0"))?

# 2) Incrementally increase the axis ratio of the lens's mass profile to 1.0. What happens to quadruple imaging?

# 3) Now, finally, think - how is all of this going to help us actually model lenses? We've said we're going to
#    reconstruct our source model_galaxy's on the pixel-grid. So, how does knowing how each pixel maps to the regular
#    actually help us? If you've not got any bright ideas, then worry not - that exactly what we're going to cover
#    in the next tutorial.


from matplotlib import pyplot as plt

from autolens.data.array.plotters import plotter_util, array_plotters
from autolens.model.profiles.plotters import profile_plotters


def plot_intensities(galaxy, grid, mask=None, positions=None, as_subplot=False,
                     units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                     cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                     cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                     title='Galaxy Intensities', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                     mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                     output_path=None, output_format='show', output_filename='galaxy_intensities'):

    intensities = galaxy.intensities_from_grid(grid=grid)
    intensities = grid.scaled_array_from_array_1d(intensities)

    array_plotters.plot_array(array=intensities, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_surface_density(galaxy, grid, mask=None, positions=None, as_subplot=False,
                         units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                         cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                         cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                         title='Galaxy Surface Density', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                         mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                         output_path=None, output_format='show', output_filename='galaxy_surface_density'):

    surface_density = galaxy.surface_density_from_grid(grid=grid)
    surface_density = grid.scaled_array_from_array_1d(surface_density)

    array_plotters.plot_array(array=surface_density, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_potential(galaxy, grid, mask=None, positions=None, as_subplot=False,
                   units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                   cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                   cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                   title='Galaxy Potential', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                   mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                   output_path=None, output_format='show', output_filename='galaxy_potential'):

    potential = galaxy.potential_from_grid(grid=grid)
    potential = grid.scaled_array_from_array_1d(potential)

    array_plotters.plot_array(array=potential, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_deflections_y(galaxy, grid, mask=None, positions=None, as_subplot=False,
                       units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       title='Galaxy Deflections (y)', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                       mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                       output_path=None, output_format='show', output_filename='galaxy_deflections_y'):

    deflections = galaxy.deflections_from_grid(grid)
    deflections_y = grid.scaled_array_from_array_1d(deflections[:,0])

    array_plotters.plot_array(array=deflections_y, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_deflections_x(galaxy, grid, mask=None, positions=None, as_subplot=False,
                       units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       title='Galaxy Deflections (x)', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                       mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                       output_path=None, output_format='show', output_filename='galaxy_deflections_x'):

    deflections = galaxy.deflections_from_grid(grid)
    deflections_x = grid.scaled_array_from_array_1d(deflections[:,1])

    array_plotters.plot_array(array=deflections_x, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_intensities_subplot(galaxy, grid, mask=None, positions=None,
                             units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                             cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                             cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                             titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                             mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                             output_path=None, output_format='show', output_filename='galaxy_individual_intensities'):

    total_light_profiles = len(galaxy.light_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=total_light_profiles)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, light_profile in enumerate(galaxy.light_profiles):

        plt.subplot(rows, columns, i+1)

        profile_plotters.plot_intensities(light_profile=light_profile, mask=mask, positions=positions, grid=grid,
                                          as_subplot=True,
                                          units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                          cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                                          linthresh=linthresh, linscale=linscale,
                                          cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                          title='Galaxy Component', titlesize=titlesize, xlabelsize=xlabelsize,
                                          ylabelsize=ylabelsize, xyticksize=xyticksize,
                                          mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                                          grid_pointsize=grid_pointsize,
                                          output_path=output_path, output_format=output_format,
                                          output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)
    plt.close()
    
def plot_surface_density_subplot(galaxy, grid, mask=None, positions=None,
                                 units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                                 cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05,
                                 linscale=0.01,
                                 cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                                 titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                                 mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                                 output_path=None, output_format='show',
                                 output_filename='galaxy_individual_surface_density'):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=total_mass_profiles)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i+1)

        profile_plotters.plot_surface_density(mass_profile=mass_profile, grid=grid, mask=mask, positions=positions,
                                              as_subplot=True,
                                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize,
                                              aspect=aspect,
                                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                                              linthresh=linthresh, linscale=linscale,
                                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                              title='Galaxy Component', titlesize=titlesize, xlabelsize=xlabelsize,
                                              ylabelsize=ylabelsize, xyticksize=xyticksize,
                                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                                              grid_pointsize=grid_pointsize,
                                              output_path=output_path, output_format=output_format,
                                              output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)
    plt.close()
    
def plot_potential_subplot(galaxy, grid, mask=None, positions=None,
                           units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                           cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                           cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                           titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                           mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                           output_path=None, output_format='show',
                           output_filename='galaxy_individual_potential'):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=total_mass_profiles)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i+1)

        profile_plotters.plot_potential(mass_profile=mass_profile, grid=grid, mask=mask, positions=positions,
                                        as_subplot=True,
                                        units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                        linscale=linscale,
                                        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                        title='Galaxy Component', titlesize=titlesize, xlabelsize=xlabelsize,
                                        ylabelsize=ylabelsize, xyticksize=xyticksize,
                                        mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                                        grid_pointsize=grid_pointsize,
                                        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)
    plt.close()
    
def plot_deflections_y_subplot(galaxy, grid, mask=None, positions=None,
                               units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                               cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05,
                               linscale=0.01,
                               cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                               titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                               mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                               output_path=None, output_format='show',
                               output_filename='galaxy_individual_deflections_y'):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=total_mass_profiles)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, mass_profile in enumerate(galaxy.mass_profiles):
    
        plt.subplot(rows, columns, i+1)
    
        profile_plotters.plot_deflections_y(mass_profile=mass_profile, grid=grid, mask=mask, positions=positions,
                                            as_subplot=True,
                                            units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                            cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                                            linthresh=linthresh, linscale=linscale,
                                            cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                            title='Galaxy Component', titlesize=titlesize, xlabelsize=xlabelsize,
                                            ylabelsize=ylabelsize, xyticksize=xyticksize,
                                            mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                                            grid_pointsize=grid_pointsize,
                                            output_path=output_path, output_format=output_format,
                                            output_filename=output_filename)
    
    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)
    plt.close()


def plot_deflections_x_subplot(galaxy, grid, mask=None, positions=None,
                               units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                               cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05,
                               linscale=0.01,
                               cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                               titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                               mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                               output_path=None, output_format='show',
                               output_filename='galaxy_individual_deflections_x'):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=total_mass_profiles)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.plot_deflections_x(mass_profile=mass_profile, grid=grid, mask=mask, positions=positions,
                                            as_subplot=True,
                                            units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                            cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                                            linthresh=linthresh, linscale=linscale,
                                            cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                            title='Galaxy Component', titlesize=titlesize, xlabelsize=xlabelsize,
                                            ylabelsize=ylabelsize, xyticksize=xyticksize,
                                            mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                                            grid_pointsize=grid_pointsize,
                                            output_path=output_path, output_format=output_format,
                                            output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)
    plt.close()


import inspect
from functools import wraps

import numpy as np


def transform_grid(func):
    """
    Wrap the function in a function that checks whether the coordinates have been transformed. If they have not been \
    transformed then they are transformed.

    Parameters
    ----------
    func : (profiles, *args, **kwargs) -> Object
        A function that requires transformed coordinates

    Returns
    -------
        A function that can except cartesian or transformed coordinates
    """

    @wraps(func)
    def wrapper(profile, grid, *args, **kwargs):
        """

        Parameters
        ----------
        profile : GeometryProfile
            The profiles that owns the function
        grid : ndarray
            PlaneCoordinates in either cartesian or profiles coordinate system
        args
        kwargs

        Returns
        -------
            A value or coordinate in the same coordinate system as those passed in.
        """
        if not isinstance(grid, TransformedGrid):
            return func(profile, profile.transform_grid_to_reference_frame(grid), *args, **kwargs)
        else:
            return func(profile, grid, *args, **kwargs)

    return wrapper


class TransformedGrid(np.ndarray):
    pass


class GeometryProfile(object):

    def __init__(self, centre=(0.0, 0.0)):
        """Abstract GeometryProfile, describing an object with y, x cartesian coordinates"""
        self.centre = centre

    def transform_grid_to_reference_frame(self, grid):
        raise NotImplemented()

    def transform_grid_from_reference_frame(self, grid):
        raise NotImplemented()

    @classmethod
    def from_profile(cls, profile, **kwargs):
        """ Creates any profiles from any other profiles, keeping all attributes from the original profiles that can
        then be passed into the constructor of the new profiles. Any none optional attributes required by the new \
        profiles's constructor which are not available as attributes of the original profiles must be passed in as \
        key word arguments. Arguments matching attributes in the original profiles may be passed in to override \
        those attributes.

        Examples
        ----------
        p = profiles.Profile(origin=(1, 1))
        elliptical_profile = profiles.EllipticalProfile.from_profile(p, axis_ratio=1, phi=2)

        elliptical_profile = profiles.EllipticalProfile(1, 2)
        profiles.Profile.from_profile(elliptical_profile).__class__ == profiles.Profile

        Parameters
        ----------
        profile: GeometryProfile
            A child of the profiles class
        kwargs
            Key word constructor arguments for the new profiles
        """
        arguments = vars(profile)
        arguments.update(kwargs)
        init_args = inspect.getfullargspec(cls.__init__).args
        arguments = {argument[0]: argument[1] for argument in arguments.items() if argument[0] in init_args}
        return cls(**arguments)

    def __repr__(self):
        return '{}\n{}'.format(self.__class__.__name__,
                               '\n'.join(["{}: {}".format(k, v) for k, v in self.__dict__.items()]))


class SphericalProfile(GeometryProfile):

    def __init__(self, centre=(0.0, 0.0)):
        """ Generic circular profiles class to contain functions shared by light and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profile.
        """
        super(SphericalProfile, self).__init__(centre)

    @transform_grid
    def grid_to_radius(self, grid):
        """
        Convert coordinates to a circular radius.

        If the coordinates have not been transformed to the profile's origin, this is performed automatically.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the profile.
        """
        return np.sqrt(np.add(np.square(grid[:, 1]), np.square(grid[:, 0])))

    def grid_angle_to_profile(self, grid_thetas):
        """The angle between each (y,x) coordinate on the grid and the profile, in radians.
        
        Parameters
        -----------
        grid_thetas : ndarray
            The angle theta counter-clockwise from the positive x-axis to each coordinate in radians.
        """
        return np.cos(grid_thetas), np.sin(grid_thetas)

    def grid_radius_to_cartesian(self, grid, radius):
        """
        Convert a grid of (y,x) coordinates with their specified circular radii to their original (y,x) Cartesian 
        coordinates.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the profile.
        radius : ndarray
            The circular radius of each coordinate from the profile center.
        """
        grid_thetas = np.arctan2(grid[:, 0], grid[:, 1])
        cos_theta, sin_theta = self.grid_angle_to_profile(grid_thetas)
        return np.multiply(radius[:, None], np.vstack((sin_theta, cos_theta)).T)

    def transform_grid_to_reference_frame(self, grid):
        """Transform a grid of (y,x) coordinates to the reference frame of the profile, including a translation to \
        its origin.

        Parameters
        ----------
        grid : ndarray
            The (y, x) coordinates in the original reference frame of the grid.
        """
        transformed = np.subtract(grid, self.centre)
        return transformed.view(TransformedGrid)

    def transform_grid_from_reference_frame(self, grid):
        """Transform a grid of (y,x) coordinates from the reference frame of the profile to the original observer \
        reference frame, including a translation from the profile's origin.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the profile.
        """
        transformed = np.add(grid, self.centre)
        return transformed.view(TransformedGrid)


class EllipticalProfile(SphericalProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """ Generic elliptical profiles class to contain functions shared by light and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalProfile, self).__init__(centre)
        self.axis_ratio = axis_ratio
        self.phi = phi

    @property
    def phi_radians(self):
        return np.radians(self.phi)

    @property
    def cos_phi(self):
        return self.cos_and_sin_from_x_axis()[0]

    @property
    def sin_phi(self):
        return self.cos_and_sin_from_x_axis()[1]

    def cos_and_sin_from_x_axis(self):
        """ Determine the sin and cosine of the angle between the profile's ellipse and the positive x-axis, \
        counter-clockwise. """
        phi_radians = np.radians(self.phi)
        return np.cos(phi_radians), np.sin(phi_radians)

    def grid_angle_to_profile(self, grid_thetas):
        """The angle between each angle theta on the grid and the profile, in radians.

        Parameters
        -----------
        grid_thetas : ndarray
            The angle theta counter-clockwise from the positive x-axis to each coordinate in radians.
        """
        theta_coordinate_to_profile = np.add(grid_thetas, - self.phi_radians)
        return np.cos(theta_coordinate_to_profile), np.sin(theta_coordinate_to_profile)

    def rotate_grid_from_profile(self, grid_elliptical):
        """ Rotate a grid of elliptical (y,x) coordinates from the reference frame of the profile back to the
        unrotated coordinate grid reference frame (coordinates are not shifted back to their original origin).

        This routine is used after computing deflection angles in the reference frame of the profile, so that the
        deflections can be re-rotated to the frame of the original coordinates before performing ray-tracing.

        Parameters
        ----------
        grid_elliptical : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of an elliptical profile.
        """
        y = np.add(np.multiply(grid_elliptical[:, 1], self.sin_phi), np.multiply(grid_elliptical[:, 0], self.cos_phi))
        x = np.add(np.multiply(grid_elliptical[:, 1], self.cos_phi), - np.multiply(grid_elliptical[:, 0], self.sin_phi))
        return np.vstack((y, x)).T

    @transform_grid
    def grid_to_elliptical_radii(self, grid):
        """
        Convert a grid of (y,x) coordinates to an elliptical radius.

        If the coordinates have not been transformed to the profile's geometry, this is performed automatically.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the elliptical profile.
        """
        return np.sqrt(np.add(np.square(grid[:, 1]), np.square(np.divide(grid[:, 0], self.axis_ratio))))

    @transform_grid
    def grid_to_eccentric_radii(self, grid):
        """
        Convert a grid of (y,x) coordinates to an eccentric radius, which is (1.0/axis_ratio) * elliptical radius and
        used to define light profile half-light radii using circular radii.

        If the coordinates have not been transformed to the profile's geometry, this is performed automatically.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the elliptical profile.
        """
        return np.multiply(np.sqrt(self.axis_ratio), self.grid_to_elliptical_radii(grid)).view(np.ndarray)

    def transform_grid_to_reference_frame(self, grid):
        """Transform a grid of (y,x) coordinates to the reference frame of the profile, including a translation to \
        its origin and a rotation to it orientation.

        Parameters
        ----------
        grid : ndarray
            The (y, x) coordinates in the original reference frame of the grid.
        """
        if self.__class__.__name__.startswith("Spherical"):
            return super().transform_grid_to_reference_frame(grid)
        shifted_coordinates = np.subtract(grid, self.centre)
        radius = np.sqrt(np.sum(shifted_coordinates ** 2.0, 1))
        theta_coordinate_to_profile = np.arctan2(shifted_coordinates[:, 0],
                                                 shifted_coordinates[:, 1]) - self.phi_radians
        transformed = np.vstack(
            (radius * np.sin(theta_coordinate_to_profile), radius * np.cos(theta_coordinate_to_profile))).T
        return transformed.view(TransformedGrid)

    def transform_grid_from_reference_frame(self, grid):
        """Transform a grid of (y,x) coordinates from the reference frame of the profile to the original observer \
        reference frame, including a rotation to its original orientation and a translation from the profile's origin.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the profile.
        """
        if self.__class__.__name__.startswith("Spherical"):
            return super().transform_grid_from_reference_frame(grid)

        y = np.add(np.add(np.multiply(grid[:, 1], self.sin_phi), np.multiply(grid[:, 0], self.cos_phi)), self.centre[0])
        x = np.add(np.add(np.multiply(grid[:, 1], self.cos_phi), - np.multiply(grid[:, 0], self.sin_phi)),
                   self.centre[1])
        return np.vstack((y, x)).T

    def eta_u(self, u, coordinates):
        return np.sqrt((u * ((coordinates[1] ** 2) + (coordinates[0] ** 2 / (1 - (1 - self.axis_ratio ** 2) * u)))))
